package Mentat::Processor::Engine::ReporterNG;
use strict;
use warnings;

#*******************************************************************************
# DOCUMENTATION SECTION
#*******************************************************************************

=head1 NAME

Mentat::Processor::Engine::ReporterNG

=head1 SYNOPSIS

 use Mentat::Processor::Engine::ReporterNG;

=head1 DESCRIPTION

Implementation of processing engine performing event reporting.

=head1 AUTHOR

 Jan Mach
 Cesnet, z.s.p.o
 jan.mach@cesnet.cz
 http://www.cesnet.cz

=head1 COPYRIGHT

This program is free software; you can redistribute
it and/or modify it under the same terms as Perl itself.

The full text of the license can be found in the
LICENSE file included with this module.

=head1 SEE ALSO

perl(1), Mentat::Processor::Engine(3), Mentat::Message::Grouper(3), Data::Filter(3),
Mentat::Stats::Data(3), Mentat::Storage::Mongo(3), Whois::Dump(3), Mentat::Message::Formater::TextTemplate(3).

=head1 FUNCTION REFERENCE

=over 4

=cut

#*******************************************************************************
# LIBRARY LOADING SECTION
#*******************************************************************************

#-- Perl core modules ---------------------------------------------------------#
use Carp;
use Scalar::Util 'blessed';
use Text::Wrap;
use Number::Format qw(:subs :vars);

use Data::Dumper; #-+-> DEVEL ONLY <-+-#
#use Smart::Comments; #-+-> DEVEL ONLY <-+-#

#-- Perl CPAN modules ---------------------------------------------------------#

#-- Custom application modules ------------------------------------------------#
use Log::Loger;
use Mentat::Core qw(:c_general :c_hawat);
use Mentat::Processor::Engine;
use Mentat::Reporter;
use Mentat::Storage::Mongo;

#*******************************************************************************
# CONSTANTS AND GLOBAL VARIABLES DEFINITION SECTION
#*******************************************************************************

#-- Constants -----------------------------------------------------------------#
use constant DFLT_DATABASE       => 'mentat';
use constant DFLT_LOG_COLLECTION => 'runlogs_reporter_ng';

#-- Static public class variables (our) ---------------------------------------#

#-- Static protected class variables (my) -------------------------------------#

#*******************************************************************************
# INITIALIZATION AND CLEANUP SECTION
#*******************************************************************************

#-- Module initializations ----------------------------------------------------#
BEGIN {
    use vars qw($VERSION $DEVEL @ISA);
    $VERSION = '0.6';
    $DEVEL   = 0;
    @ISA     = ('Mentat::Processor::Engine');
}


#-- Module clean-up code (global destructor) ----------------------------------#
END {

}

#*******************************************************************************
# IMPLEMENTATION SECTION
#*******************************************************************************

=item eh_process($) [PUBLIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Croaks, if invoked on object

=cut

sub eh_process($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($last_time,) = @_;
    Log::Loger->notice(__PACKAGE__, "[ENG:ReporterNG] Initiating reporting process...");

    # Determine the state of the last run
    my $lrl = $self->_load_pstate();
    Log::Loger->notice(__PACKAGE__, "[ENG:ReporterNG] LAST RUN STATE: ".Dumper($lrl));

    # Perform the reporting
    my ($rl, $reports) = $self->{REPORTER}->report(last_run_log => $lrl);
    #my ($rl, $reports) = $self->{REPORTER}->report_legacy(last_run_log => $lrl);
    Log::Loger->notice(__PACKAGE__, "[ENG:ReporterNG] CURRENT RUN LOG: ".Dumper($rl));

    # Update the persistent state
    $self->_update_pstate($rl);

    # Store the processing statistics to database
    $self->_runlog_store($rl);

    # Store the processing statistics to database
    $self->_runlog_log($rl);

    # Set the runlog, it will be dumped automatically after the processing is done
    $self->runlog_set($rl);

    Log::Loger->notice(__PACKAGE__, "[ENG:ReporterNG] Reporting process done...");
    return (EC_CONTINUE);
}

=item pstate($;$) [PUBLIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Dies, if not invoked on object instance

=cut

sub pstate($;$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    return $self->component('pstate')->pstate(@_);
}

=item pstate_set($$) [PUBLIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Dies, if not invoked on object instance

=cut

sub pstate_set($$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    return $self->component('pstate')->pstate_set(@_);
}

=item pstate_del($) [PUBLIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Dies, if not invoked on object instance

=cut

sub pstate_del($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    return $self->component('pstate')->pstate_del(@_);
}

=item pstate_list() [PUBLIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Dies, if not invoked on object instance

=cut

sub pstate_list()
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    return $self->component('pstate')->pstate_list();
}

=item shutdown(;$) [PUBLIC]

  Usage     : $instance->shutdown();
  Purpose   : Callback to allow the component perform arbitrary shutdown operations
  Arguments : HASH REFERENCE $config - shutdown configurations [OPTIONAL]
  Returns   : Nothing
  Throws    : Dies, if not invoked on instance or on error

=cut

sub shutdown(;$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($config,) = @_;
    confess "Component shutdown configurations must be passed via HASH reference" if $config and ref $config ne 'HASH';

    $self->_cleanup_pstate();
}

#-------------------------------------------------------------------------------
# Private interface
#-------------------------------------------------------------------------------

# _init [PROTECTED]
#
#  Usage     : return $self->_init(@_);
#  Purpose   : Initialize the new Mentat::Processor::Engine::Reporter instance
#  Arguments :
#  Returns   : Mentat::Processor::Engine::Reporter reference
#  Throws    : Dies, if not invoked on object instance
#  Comments  : This method must never be called directly, it is supposed to be overloaded by subclasses.

sub _init
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my %config = @_;

    # Call parent version first
    $self = $self->SUPER::_init(%config);

    # Initialize the reporter engine
    $self->{REPORTER} = Mentat::Reporter->new(%config);
    $self->{LSTORAGE} = Mentat::Storage::Mongo->new(database => DFLT_DATABASE(), collection => DFLT_LOG_COLLECTION());

    return $self;
}

# _cleanup_pstate() [PROTECTED]
#
#  Usage     : $self->_cleanup_pstate();
#  Purpose   : Cleanup persistent state storage
#  Arguments : None
#  Returns   : Nothing
#  Throws    : Dies, if not invoked on object instance

sub _cleanup_pstate()
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);

    # Get the list of keys in persistent state storage
    my @keys = sort $self->pstate_list();

    Log::Loger->info(__PACKAGE__, "[ENG:ReporterNG] Persistent state storage now contains following keys: ".join(',', sort $self->pstate_list()));
}

# _get_components_linked(;@) [PROTECTED]
#
#  Usage     : Used as a hook method from template method register()
#  Purpose   : Return list of required components
#  Arguments : LIST OF ARRAY REFERENCES @lists - linked component lists from parent classess [OPTIONAL]
#  Returns   : ARRAY REFERENCE $components containing STRINGS with following format:
#                  'type'      - request for component of given type and any name
#                  ':name'     - request for component of any type but given name
#                  'type:name' - request for component of given both type and name
#  Throws    : Dies, if not invoked on instance

sub _get_components_linked(;@)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my @lists = @_;

    return $self->SUPER::_get_components_linked([
        ['state', 'pstate'],
    ], @lists);
}

# _update_pstate($) [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    : Dies, if not invoked on object instance

sub _update_pstate($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($runlog,) = @_;

    # Save the state of the current run
    if ($runlog->{severities} and ref $runlog->{severities} eq 'HASH') {
        my $lth;
        foreach my $sev (keys %{$runlog->{severities}}) {
            $lth = $runlog->{severities}->{$sev}->{ts_to};
            if ($lth) {
                Log::Loger->notice(__PACKAGE__, "[ENG:ReporterNG] Updating pstate for severity '$sev' with '$lth'");
                $self->pstate_set('last_ts_'.lc($sev), $lth);
            }
        }
    }
}

# _load_pstate() [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    : Dies, if not invoked on object instance

sub _load_pstate()
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);

    # Determine the state of the last run
    my $lrl = {};
    my @keys = sort $self->pstate_list();
    foreach my $k (@keys) {
        $lrl->{$k} = $self->pstate($k);
    }
    return $lrl;
}

# _runlog_store($) [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    : Dies, if not invoked on object instance

sub _runlog_store($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($runlog,) = @_;

    eval {
        $runlog->{ts}   = time();
        $runlog->{pid}  = $$;

        $runlog->{list_analyzers}  = [sort keys %{$runlog->{analyzers}}];
        $runlog->{list_categories} = [sort keys %{$runlog->{categories}}];

        Log::Loger->debug(__PACKAGE__, "[ENG:ReporterNG] Storing run log: ".Dumper($runlog));

        my $id = $self->{LSTORAGE}->insert($runlog);
        Log::Loger->notice(__PACKAGE__, "[ENG:ReporterNG] Stored run log to MongoDB document '$id'");
    };
    if($@) {
        Log::Loger->error(__PACKAGE__, "[ENG:ReporterNG] Unable to store run log to MongoDB: $@");
    }
}

# _runlog_log($) [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    : Dies, if not invoked on object instance

sub _runlog_log($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($runlog,) = @_;

    # Format message anout the processing summary
    $Text::Wrap::columns = 100;
    my $results = '';
    $results .= "    * Total number of prefetched events:     ".format_number($runlog->{events_all} || 0)."\n";
    $results .= "    * Total number of prefiltered events:    ".format_number($runlog->{events_blk} || 0)."\n";
    $results .= "    * Total number of processed events:      ".format_number($runlog->{events_flt} || 0)."\n";
    $results .= "    * Processing time:                       ".$self->duration('AS_STRING')."\n";
    $results .= "    * Fetching time:                         ".Value::Convertor->duration_to_string($runlog->{duration_fetch})."\n"     if $runlog->{duration_fetch};
    $results .= "    * Prefiltering time:                     ".Value::Convertor->duration_to_string($runlog->{duration_prefilter})."\n" if $runlog->{duration_prefilter};
    $results .= "    * Fetching speed:                        ".format_number($runlog->{speed_fetch})." #/s\n";
    $results .= "    * Prefiltering speed:                    ".format_number($runlog->{speed_prefilter})." #/s\n";
    $results .= "    * Total number of generated reports:     ".format_number($runlog->{reports_all}         || 0)."\n";
    $results .= "    * Number of 'summary' reports:           ".format_number($runlog->{reports_t_summary}   || 0)."\n";
    $results .= "    * Number of 'extra' reports:             ".format_number($runlog->{reports_t_extra}     || 0)."\n";
    $results .= "    * Number of 'low' severity reports:      ".format_number($runlog->{reports_s_low}       || 0)."\n";
    $results .= "    * Number of 'medium' severity reports:   ".format_number($runlog->{reports_s_medium}    || 0)."\n";
    $results .= "    * Number of 'high' severity reports:     ".format_number($runlog->{reports_s_high}      || 0)."\n";
    $results .= "    * Number of 'critical' severity reports: ".format_number($runlog->{reports_s_critical}  || 0)."\n";
    $results .= "    * Number of reports archived:            ".format_number($runlog->{reports_a_archived}  || 0)."\n";
    $results .= "    * Number of reports JSON archived:       ".format_number($runlog->{reports_a_jarchived} || 0)."\n";
    $results .= "    * Number of reports filed:               ".format_number($runlog->{reports_a_filed}     || 0)."\n";
    $results .= "    * Number of reports mailed:              ".format_number($runlog->{reports_a_mailed}    || 0)."\n";
    $results .= "    * Number of reports stored:              ".format_number($runlog->{reports_a_stored}    || 0)."\n";
    #$results .= "    * Number of resolved hosts:          ".format_number($runlog->{dns_resolved_host} || 0)."\n";
    #$results .= "    * Number of resolved IPs:            ".format_number($runlog->{dns_resolved_ip}   || 0)."\n";
    #$results .= "    * Number of unresolved objects:      ".format_number($runlog->{dns_unresolved}    || 0)."\n";
    #$results .= "    * Number of DNS entries cached:      ".format_number($runlog->{dns_cache_cached}  || 0)."\n";
    #$results .= "    * Number of DNS cache hits:          ".format_number($runlog->{dns_cache_hit}     || 0)."\n";
    #$results .= "    * Number of DNS cache misses:        ".format_number($runlog->{dns_cache_miss}    || 0)."\n";
    #$results .= "    * Number of DNS query timeouts:      ".format_number($runlog->{dns_timeout}       || 0)."\n";
    #$results .= "    * Number of DNS query failures:      ".format_number($runlog->{dns_failure}       || 0)."\n";
    #$results .= "    * DNS effectivity:                   ".format_number($runlog->{dns_effect}        || 0)." %\n";
    #$results .= "    * DNS cache effectivity:             ".format_number($runlog->{dns_cache_effect}  || 0)." %\n";
    if ($runlog->{reports_all}) {
        $results .= "    * List of generated reports:\n";
        $results .= wrap('        ','        ',join(", ",@{$runlog->{ids}}))."\n";
        $results .= "    * List of related abuses:\n";
        $results .= wrap('        ','        ',join(", ", map { "$_(".$runlog->{reports_abuses}->{$_}.")" } keys(%{$runlog->{reports_abuses}})))."\n";
        $results .= "    * List of email destinations:\n";
        $results .= wrap('        ','        ',join(", ", map { "$_(".$runlog->{reports_mails}->{$_}.")" } keys(%{$runlog->{reports_mails}})))."\n";
    }
    else {
        $results .= "    * No reports were generated in this period !\n";
    }
    # Set the status message
    $self->status_msg($results);
}

=pod

=back

=cut

1;
