package Mentat::Processor::Engine::Statistician;
use strict;
use warnings;

#*******************************************************************************
# DOCUMENTATION SECTION
#*******************************************************************************

=head1 NAME

Mentat::Processor::Engine::Statistician - Short module description

=head1 SYNOPSIS

 use Mentat::Processor::Engine::Statistician;

=head1 DESCRIPTION

Extensive module description

=head1 AUTHOR

 Jan Mach
 Cesnet, z.s.p.o
 jan.mach@cesnet.cz
 http://www.cesnet.cz

=head1 COPYRIGHT

This program is free software; you can redistribute
it and/or modify it under the same terms as Perl itself.

The full text of the license can be found in the
LICENSE file included with this module.

=head1 SEE ALSO

perl(1), Mentat::Cruncher(3).

=head1 FUNCTION REFERENCE

=over 4

=cut

#*******************************************************************************
# LIBRARY LOADING SECTION
#*******************************************************************************

#-- Perl core modules ---------------------------------------------------------#
use Carp;
use Scalar::Util 'blessed';

use Data::Dumper; #-+-> DEVEL ONLY <-+-#
#use Smart::Comments; #-+-> DEVEL ONLY <-+-#

#-- Perl CPAN modules ---------------------------------------------------------#

#-- Custom application modules ------------------------------------------------#
use Log::Loger;
use Mentat::Core qw(:c_general);
use Mentat::Processor::Engine;
use Mentat::Message::Grouper;
use Mentat::Stats::Data;
use Mentat::Stats::RRD;
use Whois::Service;
use Whois::Mongo;
use Whois::Negistry;

#*******************************************************************************
# CONSTANTS AND GLOBAL VARIABLES DEFINITION SECTION
#*******************************************************************************

#-- Constants -----------------------------------------------------------------#
use constant DFLT_DATABASE         => 'mentat_stats';
use constant DFLT_STATS_COLLECTION => 'statistics';
#use constant DFLT_LOG_COLLECTION   => 'logs';

#-- Static public class variables (our) ---------------------------------------#

#-- Static protected class variables (my) -------------------------------------#

#*******************************************************************************
# INITIALIZATION AND CLEANUP SECTION
#*******************************************************************************

#-- Module initializations ----------------------------------------------------#
BEGIN {
    use vars qw($VERSION $DEVEL @ISA);
    $VERSION = '0.1';
    $DEVEL   = 0;
    @ISA     = ('Mentat::Processor::Engine');
}


#-- Module clean-up code (global destructor) ----------------------------------#
END {

}

#*******************************************************************************
# IMPLEMENTATION SECTION
#*******************************************************************************

=item eh_process($) [PUBLIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Croaks, if invoked on object

=cut

sub eh_process($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($messages, $count_all, $th, $tl) = @_;

    my $count = scalar @$messages;
    $th = int($th); $tl = int($tl); $count = int($count);

    unless ($self->{GENERATE_ONLY}) {
        # Storage for general statistics
        my $runlog = {
            'ts_start' => int(time),    # Timestamp of calculation beginning
            'ts_from'  => int($tl),     # Lower threshold
            'ts_to'    => int($th),     # Higher threshold
            'window_i' => "$tl-$th",    # Window identifier (integer)
            'count'    => int($count),  # Total event count
        };

        # Calculate and store the statistics into database
        $self->_db_update_stats($messages, $runlog);
        
        # Update source specific databases with calculated values, if possible
        if ($count) {
            $self->_rrd_update_with_counts('nodename', 'Node[#]/Name', $messages, $th, $tl);
            $self->_rrd_update_with_counts('nodesw',   'Node[#]/SW',   $messages, $th, $tl);
            $self->_rrd_update_with_counts('category', 'Category',     $messages, $th, $tl);
        }
        # Or pass zeros, so that there will not be undefined value
        else {
            $self->_rrd_update_with_defaults(0, $th, $tl);
        }

        # Set the runlog, it will be dumped automatically after the processing is done
        $self->runlog_set($runlog);
    }
    else {
        Log::Loger->notice(__PACKAGE__, "[ENG:Statistician] Skipping statistics calculation phase");
    }

    # Generate the graphs
    Log::Loger->notice(__PACKAGE__, "[ENG:Statistician] Generating graphs with upper time boundary '$th'");
    $self->{STATS_RRD}->generate_graphs($th);

    return (EC_CONTINUE, $messages, $th, $tl);
}

=item eh_gengraphs($) [PUBLIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Croaks, if invoked on object

=cut

sub eh_gengraphs($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($th,) = @_;

    Log::Loger->notice(__PACKAGE__, "[ENG:Statistician] Generating graphs with upper time boundary '$th'");
    $self->{STATS_RRD}->generate_graphs($th);

    return (EC_CONTINUE, $th);
}

#-------------------------------------------------------------------------------
# Private interface
#-------------------------------------------------------------------------------

# _init [PROTECTED]
#
#  Usage     : return $self->_init(@_);
#  Purpose   : Initialize the new Mentat::Processor::Engine::Statistician instance
#  Arguments :
#  Returns   : Mentat::Processor::Engine::Statistician reference
#  Throws    : Croaks, if invoked on class
#  Comments  : This method must never be called directly, it is supposed to be overloaded by subclasses.

sub _init
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my %config = @_;

    # Call parent version first
    $self = $self->SUPER::_init(%config);

    # Perform some sanity checking first
    die "[ENG:Statistician] Report directory must be specified under configuration key 'reports_dir'\n" unless exists $config{reports_dir};
    die "[ENG:Statistician] Report directory '$config{reports_dir}' does not exist\n" unless -d $config{reports_dir};
    die "[ENG:Statistician] Report directory '$config{reports_dir}' is not writable\n" unless -w $config{reports_dir};
    die "[ENG:Statistician] Chart directory must be specified under configuration key 'charts_dir'\n" unless exists $config{charts_dir};
    die "[ENG:Statistician] Chart directory '$config{charts_dir}' does not exist\n" unless -d $config{charts_dir};
    die "[ENG:Statistician] Chart directory '$config{charts_dir}' is not writable\n" unless -w $config{charts_dir};
    die "[ENG:Statistician] RRD database directory must be specified under configuration key 'rrds_dir'\n" unless exists $config{rrds_dir};
    die "[ENG:Statistician] RRD database directory '$config{rrds_dir}' does not exist\n" unless -d $config{rrds_dir};
    die "[ENG:Statistician] RRD database directory '$config{rrds_dir}' is not writable\n" unless -w $config{rrds_dir};
    die "[ENG:Statistician] Whois database name must be specified under configuration key 'whois_database_name'\n" unless exists $config{whois_database_name};
    die "[ENG:Statistician] Whois collection name must be specified under configuration key 'whois_collection_name'\n" unless exists $config{whois_collection_name};

    $self->{GROUPER}   = Mentat::Message::Grouper->new();
    $self->{STATS}     = Mentat::Stats::Data->new(charts_dir => $config{charts_dir});
    $self->{STATS_RRD} = Mentat::Stats::RRD->new(reports_dir => $config{reports_dir},
                                                 rrds_dir => $config{rrds_dir},
                                                 colors => $config{colors},
                                                 titles => $config{titles},
                         );

    $self->{WHOIS} = Whois::Service->new(
                            Whois::Mongo->new($config{whois_database_name}, $config{whois_collection_name}),
                            #Whois::Negistry->new($config{whois_dump_file}, 1, $config{whois_exceptions}),
                         );

    $self->{STORAGE} = Mentat::Storage::Mongo->new(database => DFLT_DATABASE(), collection => DFLT_STATS_COLLECTION());
    #$self->{LSTORAGE} = Mentat::Storage::Mongo->new(database => DFLT_DATABASE(), collection => DFLT_LOG_COLLECTION());

    $self->{GENERATE_ONLY} = $config{generate_only};

    Log::Loger->info(__PACKAGE__, "[ENG:Statistician] Using folder '$config{charts_dir}' as chart destination");
    Log::Loger->info(__PACKAGE__, "[ENG:Statistician] Using folder '$config{reports_dir}' as report destination");
    Log::Loger->info(__PACKAGE__, "[ENG:Statistician] Using folder '$config{rrds_dir}' as RRD database location");

    return $self;
}

# _get_events_handled(;@) [PROTECTED]
#
#  Usage     : Used as a hook method from template method register()
#  Purpose   : Return list of events and appropriate handlers for this component
#  Arguments : LIST OF ARRAY REFERENCES @lists - event handler lists from parent classess [OPTIONAL]
#  Returns   : ARRAY REFERENCE $handlers with internal structure [$event, $handler]
#                  STRING $event   - event name
#                  STRING $handler - event handler name
#  Throws    : Dies, if not invoked on instance

sub _get_events_handled(;@)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my @lists = @_;

    return $self->SUPER::_get_events_handled([
            ['messages_fetched', 'eh_process'],
        ], @lists);
}

# _rrd_update_with_counts($$$$$) [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    : Croaks, if invoked on class

sub _rrd_update_with_counts($$$$$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($type, $grouping, $messages, $th, $tl) = @_;

    # Get the total count
    my $count = scalar @$messages;
    Log::Loger->notice(__PACKAGE__, "[ENG:Statistician] Updating RRD DB '$type.".Mentat::Stats::RRD::DB_TOTALS_NAME."' statistics with value '$count' and timestamp '$th'");
    $self->{STATS_RRD}->update($type.'.'.Mentat::Stats::RRD::DB_TOTALS_NAME, $count, $th);

    # Find RRD database files with given type of data
    my $dbs_r = $self->{STATS_RRD}->find_dbs($type, Mentat::Stats::RRD::DB_TOTALS_NAME);

    my %dbs_e = map { $_->[0] => 1 } ((ref $dbs_r->{$type})?@{$dbs_r->{$type}}:());

    my ($value, $source);
    $messages = $self->{GROUPER}->group($grouping, $messages);
    foreach $source (keys(%{$messages}))    {
        $value  = scalar @{$messages->{$source}};
        $source = _clean_source($source);

        Log::Loger->notice(__PACKAGE__, "[ENG:Statistician] Updating RRD DB '$type.$source' statistics with value '$value' and timestamp '$th'");
        $self->{STATS_RRD}->update("$type.$source", $value, $th);

        delete $dbs_e{"$type.$source"};
    }

    foreach my $source_id (keys(%dbs_e))    {
        Log::Loger->notice(__PACKAGE__, "[ENG:Statistician] Updating RRD DB '$source_id' statistics with default value '0' and timestamp '$th'");
        $self->{STATS_RRD}->update("$source_id", 0, $th);
    }
}

# _rrd_update_with_defaults($) [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    : Croaks, if invoked on class

sub _rrd_update_with_defaults($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($value, $th, $tl) = @_;

    Log::Loger->notice(__PACKAGE__, "[ENG:Statistician] Updating all existing RRD DBs with default value '$value' and timestamp '$th'");
    $self->{STATS_RRD}->update_all($value, $th);
}

# _db_update_stats($) [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    : Croaks, if invoked on class

sub _db_update_stats($$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($messages, $stats) = @_;

    $stats->{Mentat::Stats::Data::ST_OVERALL()} = $self->{STATS}->evaluate_alerts($messages, {}, 1);

    my $rule  = [\&_agr_by_abuse, 'Source', $self->{WHOIS}];
    my $gmsgs = $self->{GROUPER}->group($rule, $messages);
    my $s;
    foreach my $type (sort keys %$gmsgs) {
        $stats->{$type} = $self->{STATS}->evaluate_alerts($gmsgs->{$type}, {}, 1);
    }

    # Add or edit some items in stats log
    $stats->{'ts'}       = int(time);                               # Timestamp of calculation end
    $stats->{'duration'} = int($stats->{ts} - $stats->{ts_start});  # Calculation duration

    Log::Loger->notice(__PACKAGE__, "[ENG:Statistician] Storing statistics log: ".Dumper($stats)) if $DEVEL;
    eval {
        my $id = $self->{STORAGE}->insert($stats);
        Log::Loger->notice(__PACKAGE__, "[ENG:Statistician] Stored statistics log to MongoDB document '$id'");

        $stats->{'flag_stored'}++;
    };
    if($@) {
        Log::Loger->error(__PACKAGE__, "[ENG:Statistician] Unable to store statistics log to MongoDB: '$@'");
        Log::Loger->error(__PACKAGE__, "[ENG:Statistician] Log dump: ".Dumper($stats));
        
    }
}

# _clean_source($) [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    : Croaks, if invoked on class

sub _clean_source($)
{
    my $s = shift;
    $s =~ s/[-+=(){}\[\]!?*%#~^&\\\/\$\@. ]/_/g;
    $s = '_unknown' unless $s;
    return $s;
}


# _agr_by_abuse($$) [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    :

sub _agr_by_abuse($$)
{
    my ($msg, $type, $whois) = @_;
    confess "No or invalid type given as argument" unless ($type and ($type =~ /Source|Target/));

    my $ips_src = $msg->path_values("$type/IP4");

    my $keys = {};
    foreach my $ip (@$ips_src) {
        my $abuse = $whois->find_abuses($ip);
        if (defined $abuse) {
            $keys->{Mentat::Stats::Data::ST_INTERNAL()}++;
        }
    }

    my @keys = keys(%$keys);
    return [Mentat::Stats::Data::ST_EXTERNAL()] unless scalar @keys;
    return \@keys;
}

=pod

=back

=cut

1;
