package Mentat::Stats::RRD;
use strict;
use warnings;

#*******************************************************************************
# DOCUMENTATION SECTION
#*******************************************************************************

=head1 NAME

Mentat::Stats::RRD - Short module description

=head1 SYNOPSIS

 use Mentat::Stats::RRD;

=head1 DESCRIPTION

Note: Rewritten the whole module to use RRDs library directly, because previously
used solution RRDTool::OO did not implement some features and options.

=head1 AUTHOR

 Jan Mach
 Cesnet, z.s.p.o
 jan.mach@cesnet.cz
 http://www.cesnet.cz

=head1 COPYRIGHT

This program is free software; you can redistribute
it and/or modify it under the same terms as Perl itself.

The full text of the license can be found in the
LICENSE file included with this module.

=head1 SEE ALSO

perl(1), Mentat::Stats(3).

=head1 FUNCTION REFERENCE

=over 4

=cut

#*******************************************************************************
# LIBRARY LOADING SECTION
#*******************************************************************************

#-- Perl core modules ---------------------------------------------------------#
use Carp;
use Scalar::Util 'blessed';
use DirHandle;
use File::Basename;

use Data::Dumper; #-+-> DEVEL ONLY <-+-#
#use Smart::Comments; #-+-> DEVEL ONLY <-+-#

#-- Perl CPAN modules ---------------------------------------------------------#
use Date::Format qw(time2str);
use RRDs;
use JSON;

#-- Custom application modules ------------------------------------------------#
use Log::Loger;
use Color::Palette;
use Mentat::Stats;

#*******************************************************************************
# CONSTANTS AND GLOBAL VARIABLES DEFINITION SECTION
#*******************************************************************************

#-- Constants -----------------------------------------------------------------#
use constant DFLT_STEP         => 300;
use constant DFLT_WIDTH        => 800;
use constant DFLT_HEIGHT       => 300;
use constant DFLT_SPARK_WIDTH  => 100;
use constant DFLT_SPARK_HEIGHT => 35;
use constant DS_NAME           => 'eventcnt';
use constant DB_TOTALS_NAME    => '_totals';

#-- Static public class variables (our) ---------------------------------------#

#-- Static protected class variables (my) -------------------------------------#

#*******************************************************************************
# INITIALIZATION AND CLEANUP SECTION
#*******************************************************************************

#-- Module initializations ----------------------------------------------------#
BEGIN {
    use vars qw($VERSION $DEVEL @ISA);
    $VERSION = '0.1';
    $DEVEL   = 0;
    @ISA     = ('Mentat::Stats');
}


#-- Module clean-up code (global destructor) ----------------------------------#
END {

}

#*******************************************************************************
# IMPLEMENTATION SECTION
#*******************************************************************************

=item DESTROY() [PUBLIC]

  Usage     : This method must not be used directly
  Purpose   : Class destructor
  Arguments : None
  Returns   : Nothing
  Throws    : Dies, if not invoked on instance

=cut

sub DESTROY()
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    ### Destroying Mentat::Stats::RRD instance...
}

#-------------------------------------------------------------------------------

=item find_dbs(;$$) [PUBLIC]

  Usage     : my @db_files = $instance->find_dbs();
  Purpose   : Find RRD database file
  Arguments : STRING $type - data type [OPTIONAL]
              STRING $skip - skip pattern [OPTIONAL]
  Returns   : ARRAY of LISTS containing:
                  STRING $db_id - id of the database
                  STRING $db_file - path to the database file
  Throws    : Dies, if not invoked on object instance

=cut

sub find_dbs(;$$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($type, $skip) = @_;

    # Database file name RE
    my $dbre = '^(.+)\.rrd$';

    my $rrds_dir = $self->{RRDS_DIR};
    my %rrds = ();

    my ($dir_handle, $file, $db_type, $db_name, $db_id);
    opendir($dir_handle, $rrds_dir) or die "Can't open RRD database directory '$rrds_dir': $!\n";
    while (defined($file = readdir($dir_handle))) {
        # Only plain files
        next unless (-f "$rrds_dir/$file");
        # Return only RRD databases
        next unless ($file =~ /$dbre/);

        # Mark down the untainted DB name
        $db_id = $1;
        ($db_type, $db_name) = split(/\./, $db_id, 2);
        confess "Invalid RRD database '$db_id', could not parse out the type and name" unless $db_type and $db_name;

        # Filter the requested database types
        next if $type and not $db_type eq $type;

        # Filter the skipped entries
        next if $skip and $file =~ /$skip/;

        # Push it into the result set
        push(@{$rrds{$db_type}}, [$db_id, $db_type, $db_name, "$rrds_dir/$file"]);
    }
    closedir($dir_handle);

    # Sort the results
    foreach my $t (keys %rrds) {
        $rrds{$t} = [sort { $a->[0] cmp $b->[0] } @{$rrds{$t}}];
    }

    return \%rrds;
}

=item prepare_db($$) [PUBLIC]

  Usage     : my $rrd = $instance->prepare_db($db_id, $time_end);
  Purpose   : Prepare the RRD database with given ID
  Arguments : STRING $db_id - unique identifier of the database [MANDATORY]
              INTEGER $time_start - start time
  Returns   : RRDTool::OO $rrd
  Throws    : Dies, if not invoked on object instance

=cut

sub prepare_db($$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($db_id, $time_start) = @_;

    # Generate name of the RRD database file
    my $filename = $self->{RRDS_DIR}."/$db_id.rrd";

    # Nothing to do, if the database already exists
    return $filename if -f $filename;

    # Calculate steps per hour
    my $sph = int(3600 / $self->{STEP});

    # Calculate the size for three month storage
    my $r_rows = $sph * 24 * 7 * 31 * 3;

    # Otherwise we have to create it
    my @result = RRDs::create(
        # Name of the database file
        $filename,
        # Start time
        '--start', ($time_start - $self->{STEP}),
        # Update step
        '--step', $self->{STEP},
        # Define data store
        'DS:'.DS_NAME().':ABSOLUTE:'.(2*$self->{STEP}).':U:U',
        # 3 month raw storage ($sph * 24 * 7 * 31 * 3)
        'RRA:MAX:0.5:1:'.$r_rows,
        # 1 year average storage (2 * 24 * 31 * 12)
        'RRA:AVERAGE:0.5:6:17856',
        # 1 year min storage (2 * 24 * 31 * 12)
        'RRA:MIN:0.5:6:17856',
        # 1 year max storage (2 * 24 * 31 * 12)
        'RRA:MAX:0.5:6:17856',
    );
    my $error = RRDs::error();
    die "Unable to create RRD database '$db_id' in file '$filename': $error\n" if $error;

    # Return the database manipulation object
    Log::Loger->notice(__PACKAGE__, "[MSM] Created new empty RRD DB '$db_id' in file '$filename'");
    return $filename;
}

=item generate_graphs($;$) [PUBLIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Dies, if not invoked on object instance

=cut

sub generate_graphs($;$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($time_end, $type) = @_;

    # Find all RRD databases
    my $rrds = $self->find_dbs($type);
    return 1 unless scalar keys %$rrds;

    # Make sure '_totals' is the first entry
    # FIXME: Really make sure, that '_totals' is the last entry, this is naive
    #@rrds = sort {$a->[2] cmp $b->[2]} @rrds;
    #push(@rrds, shift(@rrds));

    my $n;              # Cleaned name of the database
    my $counter = 0;    # Counter
    my $rrd_args;       # Arguments for RRD graph generation
    my @graphlist;      # List of RRD graphs to be generated
    my @sparkgraphlist; # List of Sparkline graphs to be generated

    # Generate graphs for all types of RRD databases separately
    foreach my $t (sort keys %$rrds) {
        my $length = 0;
        foreach my $rrd (@{$rrds->{$t}}) {
            $length = length($rrd->[2]) if $length < length($rrd->[2]);
        }

        Log::Loger->notice(__PACKAGE__, "[MSM] RRD type: $t");
        $rrd_args = {
                'defs'   => [], 'draws'   => [], 'xports'   => [],
                't_defs' => [], 't_draws' => [], 't_xports' => [],
            };

        # Prepare RRD args for each
        foreach my $rrd (@{$rrds->{$t}}) {
            $counter = $self->_rrd_prepare_args($t, $rrd, $rrd->[0], $rrd->[2], $counter, $rrd_args, $length);
        }

        # Get the lists of graphs to be generated (regular graphs, spark graphs)
        @graphlist      = $self->_rrd_prepare_graphs($t, $time_end, $rrd_args);
        @sparkgraphlist = $self->_rrd_prepare_sparkgraphs($t, $time_end, $rrd_args);

        # Generate the graphs
        foreach my $g (@graphlist) {
            $self->_rrd_generate_graph(@$g);
            $self->_rrd_xport(@$g);
        }
        # Generate the sparkline graphs
        foreach my $g (@sparkgraphlist) {
            $self->_rrd_generate_sparkgraph(@$g);
        }
    }
}

=item update($$;$) [PUBLIC]

  Usage     : my $rv = $instance->update($id, $value, $ts);
  Purpose   : Update given RRD database file with given value
  Arguments : STRING $rrd_id - identifier of the database [MANDATORY]
              INTEGER $value - value [MANDATORY]
              INTEGER $time - update timestamp [OPTIONAL]
  Returns   : INTEGER 1 on success, 0 on FAILURE
  Throws    : Dies, if not invoked on object instance

=cut

sub update($$;$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($rrd_id, $value, $time) = @_;

    # Setup the default timestamp, if necessary
    $time = ($time) ? $time : time();

    # Get the database manipulation object
    my $rrddb = $self->prepare_db($rrd_id, $time);

    # Update the RRD database file
    Log::Loger->info(__PACKAGE__, "[MSM] Updating RRD DB '$rrd_id' with value '$value' and timestamp '$time'");
    my @result = RRDs::update(
        $rrddb,
        "$time:$value",
    );
    my $error = RRDs::error();
    die "Unable to update RRD database '$rrd_id' in file '$rrddb': $error\n" if $error;

    return 1;
}

=item update_all($;$$$) [PUBLIC]

  Usage     : my $cnt = $instance->update($value, $ts, $type, $skip);
  Purpose   : Update particular/all RRD database files with given value
  Arguments : INTEGER $value - value [MANDATORY]
              INTEGER $time - update timestamp [OPTIONAL]
              STRING $type - data type [OPTIONAL]
              STRING $skip - skip pattern [OPTIONAL]
  Returns   : INTEGER $count number of updated database files
  Throws    : Dies, if not invoked on object instance
  See Also  : find_dbs() and update() methods

=cut

sub update_all($;$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($value, $time, $type, $skip) = @_;

    # Find all databases except the totals one
    my $rrds = $self->find_dbs($type, $skip);

    my $cnt = 0; my $rc;
    foreach my $type (keys %$rrds) {
        foreach my $rrd (@{$rrds->{$type}}) {
            $rc = $self->update($rrd->[0], $value, $time);
            $cnt++ if $rc;
        }
    }
    return $cnt;
}

#-------------------------------------------------------------------------------
# Private interface
#-------------------------------------------------------------------------------

# _init [PROTECTED]
#
#  Usage     : return $self->_init(@_);
#  Purpose   : Initialize the new Mentat::Stats::RRD instance
#  Arguments : HASH %config - instance configurations [MANDATORY]
#  Returns   : Mentat::Stats::RRD reference
#  Throws    : Dies, if not invoked on object instance
#  Comments  : This method must never be called directly, it is supposed to be overloaded by subclasses.

sub _init
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my %config = @_;
    my ($reports_dir, $rrds_dir, $colors, $step, $agregation, $width, $height) = @_;
    confess "Report and RRD database directory names are mandatory arguments" unless $config{reports_dir} and $config{rrds_dir};
    confess "Report dir must be existing directory" unless -d $config{reports_dir};
    confess "RRD database must be existing directory" unless -d $config{rrds_dir};

    $self->{REPORTS_DIR}  = $config{reports_dir};
    $self->{RRDS_DIR}     = $config{rrds_dir};
    $self->{COLORS}       = $config{colors}       || {};
    $self->{TITLES}       = $config{titles}       || {};
    $self->{STEP}         = $config{step}         || DFLT_STEP;
    $self->{WIDTH}        = $config{width}        || DFLT_WIDTH;
    $self->{HEIGHT}       = $config{height}       || DFLT_HEIGHT;
    $self->{SPARK_WIDTH}  = $config{spark_width}  || DFLT_SPARK_WIDTH;
    $self->{SPARK_HEIGHT} = $config{spark_height} || DFLT_SPARK_HEIGHT;

    die "Invalid step value $self->{STEP}, valid value is >= 5\n" unless $self->{STEP} >= 5;

    Log::Loger->notice(__PACKAGE__, "[MSM] Using predefined color palete: " . Dumper($self->{COLORS}));

    return $self;
}

# _color_palette() [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    : Dies, if not invoked on object instance

sub _color_palette()
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);

    $self->{COLOR_PALETTE} = Color::Palette->new() unless $self->{COLOR_PALETTE};

    return $self->{COLOR_PALETTE};
}

# _color_for_db($$) [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    : Dies, if not invoked on object instance

sub _color_for_db($$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($set, $db_id) = @_;

    # If the color for this database was specified explicitly, return it
    if ($self->{COLORS}->{$set}->{$db_id}) {
        Log::Loger->notice(__PACKAGE__, "[MSM] Using predefined color '" . $self->{COLORS}->{$set}->{$db_id} . "' for RRD DB '$db_id' in set '$set'");
        return $self->_color_palette->as_rgb($self->{COLORS}->{$set}->{$db_id});
    }

    # Draw the totals in red, there is always only one line in graph
    my $ptrn = DB_TOTALS_NAME().'$';
    return ('red', 'FF0000') if ($db_id =~ /$ptrn/);

    # Otherwise generate new random color and store it for this database
    my ($color_str, $color_hex) = $self->_color_palette->next_color(values(%{$self->{COLORS}->{$set}}));
    $self->{COLORS}->{$set}->{$db_id} = $color_str;

    Log::Loger->notice(__PACKAGE__, "[MSM] Color '$color_str' was chosen for RRD DB '$db_id' in set '$set'");

    # And return it
    return $color_hex;
}

# _title_for_db_type($) [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    : Dies, if not invoked on object instance

sub _title_for_db_type($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($db_type,) = @_;

    # If the color for this database was specified explicitly, return it
    if ($self->{TITLES}->{$db_type}) {
        return $self->{TITLES}->{$db_type};
    }
    return 'Events processed';
}

# _rrd_prepare_args($$$$$) [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    : Dies, if not invoked on object instance

sub _rrd_prepare_args($$$$$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($type, $rrd, $id, $n, $counter, $rrd_args, $length) = @_;

    # Get the color for data from this source
    my $color = $self->_color_for_db($type, $id);

    $id =~ s/\./_/g;
    $counter++;

    my $dsid = DS_NAME().'_'.$id;
    my $totals = 0;
    my $legend = '# of messages from '.sprintf("%-${length}s", $n);

    my $ptrn = DB_TOTALS_NAME().'$';
    if ($id =~ /$ptrn/) { $totals = 1; $legend = 'Total # of messages' }

    my @defs = (
        # 'DEF:eventcnt_proc_a01=./spool/proc.a01.rrd:eventcnt:MIN'
        'DEF:'.$dsid.'='.$rrd->[3].':'.DS_NAME().':MAX',
        # 'CDEF:eventcnt_proc_a01_r=eventcnt_proc_a01,300,*'
        'CDEF:'.$dsid.'_r='.$dsid.',300,*',
        # Calculate current value
        # 'VDEF:cur_eventcnt_proc_a01=eventcnt_proc_a01_r,LAST'
        'VDEF:cur_'.$dsid.'='.$dsid.'_r,LAST',
        # Calculate overall average value
        # 'VDEF:avg_eventcnt_proc_a01=eventcnt_proc_a01_r,AVERAGE'
        'VDEF:avg_'.$dsid.'='.$dsid.'_r,AVERAGE',
        # Calculate overall maximum value
        # 'VDEF:max_eventcnt_proc_a01=eventcnt_proc_a01_r,MAXIMUM'
        'VDEF:max_'.$dsid.'='.$dsid.'_r,MAXIMUM',
        # Calculate overall minimum value
        # 'VDEF:min_eventcnt_proc_a01=eventcnt_proc_a01_r,MINIMUM'
        'VDEF:min_'.$dsid.'='.$dsid.'_r,MINIMUM',
    );
    my @draws = (
        # Diplay the actual data as stacked area
        # 'AREA:draw10#FFFF00:# of messages processed from a02:STACK'
        'AREA:'.$dsid.'_r#'.$color.':'.$legend.(($counter>1)?':STACK':''),
        # Display current value in the labels
        # 'GPRINT:cur_eventcnt_proc_a01:Cur [a01] = %lf'
        'GPRINT:cur_'.$dsid.':Cur\: %8.1lf',
        # Display overall average in the labels
        # 'GPRINT:avg_eventcnt_proc_a01:Avg [a01] = %lf'
        'GPRINT:avg_'.$dsid.':Avg\: %8.1lf',
        # Display overall maximum in the labels
        # 'GPRINT:max_eventcnt_proc_a01:Max [a01] = %lf'
        'GPRINT:max_'.$dsid.':Max\: %8.1lf',
        # Display overall minimum in the labels
        # 'GPRINT:min_eventcnt_proc_a01:Min [a01] = %lf'
        'GPRINT:min_'.$dsid.':Min\: %8.1lf\n',
    );
    my @xports = (
        # Xport the actual data
        'XPORT:'.$dsid.'_r:'.$legend,
    );

    if ($totals) {
        push(@{$rrd_args->{t_defs}},   @defs);
        push(@{$rrd_args->{t_draws}},  @draws);
        push(@{$rrd_args->{t_xports}}, @xports);
    }
    else {
        push(@{$rrd_args->{defs}},   @defs);
        push(@{$rrd_args->{draws}},  @draws);
        push(@{$rrd_args->{xports}}, @xports);
    }

    return $counter;
}

# _rrd_prepare_graphs($$$) [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    : Dies, if not invoked on object instance

sub _rrd_prepare_graphs($$$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($rrd_type, $time_end, $rrd_args) = @_;

    my $title = $self->_title_for_db_type($rrd_type);
    return (
        # $fname, $title, $label, $start, $end, $a1, $a2
        [
            "$rrd_type.l6hours",
            $title . ' in last 6 hours',
            '[#]', $time_end - 21600, $time_end,
            $rrd_args->{defs}, $rrd_args->{draws}, $rrd_args->{xports},
        ],
        [
            "$rrd_type.l24hours",
            $title . ' in last 24 hours',
            '[#]', $time_end - 86400, $time_end,
            $rrd_args->{defs}, $rrd_args->{draws}, $rrd_args->{xports},
        ],
        [
            "$rrd_type.lweek",
            $title . ' in last week',
            '[#]', $time_end - 604800, $time_end,
            $rrd_args->{defs}, $rrd_args->{draws}, $rrd_args->{xports},
        ],
        [
            "$rrd_type.l4weeks",
            $title . ' in last 4 weeks',
            '[#]', $time_end - 2419200, $time_end,
            $rrd_args->{defs}, $rrd_args->{draws}, $rrd_args->{xports},
        ],
        [
            "$rrd_type.l3months",
            $title . ' in last 3 months',
            '[#]', $time_end - 7257600, $time_end,
            $rrd_args->{defs}, $rrd_args->{draws}, $rrd_args->{xports},
        ],
        [
            "$rrd_type.l6months",
            $title . ' in last 6 months',
            '[#]', $time_end - 14515200, $time_end,
            $rrd_args->{defs}, $rrd_args->{draws}, $rrd_args->{xports},
        ],
        [
            "$rrd_type.l12months",
            $title . ' in last 12 months',
            '[#]', $time_end - 29030400, $time_end,
            $rrd_args->{defs}, $rrd_args->{draws}, $rrd_args->{xports},
        ],
        [
            "$rrd_type.l6hours-t",
            $title . ' in last 6 hours (TOTAL)',
            '[#]', $time_end - 21600, $time_end,
            $rrd_args->{t_defs}, $rrd_args->{t_draws}, $rrd_args->{t_xports},
        ],
        [
            "$rrd_type.l24hours-t",
            $title . ' in last 24 hours (TOTAL)',
            '[#]', $time_end - 86400, $time_end,
            $rrd_args->{t_defs}, $rrd_args->{t_draws}, $rrd_args->{t_xports},
        ],
        [
            "$rrd_type.lweek-t",
            $title . ' in last week (TOTAL)',
            '[#]', $time_end - 604800, $time_end,
            $rrd_args->{t_defs}, $rrd_args->{t_draws}, $rrd_args->{t_xports},
        ],
        [
            "$rrd_type.l4weeks-t",
            $title . ' in last 4 weeks (TOTAL)',
            '[#]', $time_end - 2419200, $time_end,
            $rrd_args->{t_defs}, $rrd_args->{t_draws}, $rrd_args->{t_xports},
        ],
        [
            "$rrd_type.l3months-t",
            $title . ' in last 3 months (TOTAL)',
            '[#]', $time_end - 7257600, $time_end,
            $rrd_args->{t_defs}, $rrd_args->{t_draws}, $rrd_args->{t_xports},
        ],
        [
            "$rrd_type.l6months-t",
            $title . ' in last 6 months (TOTAL)',
            '[#]', $time_end - 14515200, $time_end,
            $rrd_args->{t_defs}, $rrd_args->{t_draws}, $rrd_args->{t_xports},
        ],
        [
            "$rrd_type.l12months-t",
            $title . ' in last 12 months (TOTAL)',
            '[#]', $time_end - 29030400, $time_end,
            $rrd_args->{t_defs}, $rrd_args->{t_draws}, $rrd_args->{t_xports},
        ],
    );
}

# _rrd_prepare_sparkgraphs($$$) [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    : Dies, if not invoked on object instance

sub _rrd_prepare_sparkgraphs($$$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($rrd_type, $time_end, $rrd_args) = @_;

    my $title = $self->_title_for_db_type($rrd_type);
    return (
        # $fname, $title, $label, $start, $end, $a1, $a2
        [
            "$rrd_type.l6hours",
            $time_end - 21600, $time_end,
            $rrd_args->{defs}, $rrd_args->{draws},
        ],
        [
            "$rrd_type.l24hours",
            $time_end - 86400, $time_end,
            $rrd_args->{defs}, $rrd_args->{draws},
        ],
        [
            "$rrd_type.lweek",
            $time_end - 604800, $time_end,
            $rrd_args->{defs}, $rrd_args->{draws},
        ],
        [
            "$rrd_type.l4weeks",
            $time_end - 2419200, $time_end,
            $rrd_args->{defs}, $rrd_args->{draws},
        ],
        [
            "$rrd_type.l6hours-t",
            $time_end - 21600, $time_end,
            $rrd_args->{t_defs}, $rrd_args->{t_draws},
        ],
        [
            "$rrd_type.l24hours-t",
            $time_end - 86400, $time_end,
            $rrd_args->{t_defs}, $rrd_args->{t_draws},
        ],
        [
            "$rrd_type.lweek-t",
            $time_end - 604800, $time_end,
            $rrd_args->{t_defs}, $rrd_args->{t_draws},
        ],
        [
            "$rrd_type.l4weeks-t",
            $time_end - 2419200, $time_end,
            $rrd_args->{t_defs}, $rrd_args->{t_draws},
        ],
    );
}

# _rrd_xport($$$$$) [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    : Dies, if not invoked on object instance
#  Comments  :
#  See Also  :

sub _rrd_xport($$$$$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($fid, $title, $label, $start, $end, $defs, $draws, $xports) = @_;

    my $fname = $self->{REPORTS_DIR}."/$fid.xport.json";
    Log::Loger->notice(__PACKAGE__, "[MSM] Exporting graph data '$fid:$title' into file '$fname'");

    my ($r_start, $r_end, $r_step, $r_cols, $r_names, $r_data) = RRDs::xport(
        '--start', $start,
        '--end', $end,
        '--json',
        @$defs,
        @$xports,
    );
    my $error = RRDs::error();
    die "Unable to export RRD graph data '$fid:$title' into file '$fname': $error\n" if $error;

    my $data = {
            'start'   => $r_start,
            'end'     => $r_end,
            'step'    => $r_step,
            'columns' => $r_cols,
            'names'   => $r_names,
            'data'    => $r_data,
        };
    my $json = to_json($data, {utf8 => 1, pretty => 1});
    open(my $hnd, ">$fname") or die "Cannot open file '$fname' for writing\n";
    print $hnd $json;
    close($hnd);
}

# _rrd_generate_graph($$$$$$$) [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    : Dies, if not invoked on object instance
#  Comments  :
#  See Also  :

sub _rrd_generate_graph($$$$$$$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($fid, $title, $label, $start, $end, $defs, $draws, $xports) = @_;

    my $fname = $self->{REPORTS_DIR}."/$fid.png";
    Log::Loger->notice(__PACKAGE__, "[MSM] Generating graph '$fid:$title' into file '$fname'");

    my ($result_arr,$xsize,$ysize) = RRDs::graph(
        $fname,
        '--width', $self->{WIDTH},
        '--height', $self->{HEIGHT},
        '--title', $title,
        '--watermark', "Generated at ".time2str('%Y-%m-%d %T%z', time)." by Mentat system, CESNET, z.s.p.o.",
        '--vertical-label', $label,
        '--right-axis-label', $label,
        '--start', $start,
        '--end', $end,
        '--slope-mode',
        @$defs,
        @$draws,
    );
    my $error = RRDs::error();
    die "Unable to generate RRD graph '$fid:$title' into file '$fname': $error\n" if $error;
}

# _rrd_generate_sparkgraph($$$$$) [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    : Dies, if not invoked on object instance
#  Comments  :
#  See Also  :

sub _rrd_generate_sparkgraph($$$$$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($fid, $start, $end, $defs, $draws) = @_;

    my $fname = $self->{REPORTS_DIR}."/$fid.spark.png";
    Log::Loger->notice(__PACKAGE__, "[MSM] Generating sparkline graph '$fid' into file '$fname'");

    my ($result_arr,$xsize,$ysize) = RRDs::graph(
        $fname,
        '--width',  100,
        '--height', 35,
        '--start',  $start,
        '--end',    $end,
        '--slope-mode',
        '--only-graph',
        '--no-legend',
        '--x-grid', 'none',
        '--y-grid', 'none',
        '--color', 'BACK#FFFFFF',
        '--color', 'ARROW#FFFFFF',
        '--color', 'CANVAS#FFFFFF',
        '--color', 'SHADEA#FFFFFF',
        '--color', 'SHADEB#FFFFFF',
        '--color', 'GRID#FFFFFF',
        '--color', 'MGRID#FFFFFF',
        '--color', 'FRAME#FFFFFF',
        '--color', 'FONT#FFFFFF',
        @$defs,
        @$draws,
    );
    my $error = RRDs::error();
    die "Unable to generate sparkline RRD graph '$fid' into file '$fname': $error\n" if $error;
}

=pod

=back

=cut

1;
